#input beam waist and cavity waist in microns, beam displacement in centimeters, wavelength in nanometers.
#laser and cavity FWHM must just be consistent units, unless you enter the finesse and FSR instead, in which case use the cavity FWHM is caculated in Hz
#beam displacement is the distance between cavity waist and actual beam waist (sign doesn't matter since it's symmetric)


import math
def spacial_overlap(beam_waist, cavity_waist, beam_dis, wavelength=1064):
    beam_waist=beam_waist*10**-6
    cavity_waist=cavity_waist*10**-6
    beam_dis=beam_dis*10**-2
    wavelength=wavelength*10**-9
    Wa_d=(beam_waist)*(1+((wavelength)*(beam_dis)/(math.pi*(beam_waist**2)))**2)**.5
    if beam_dis==0:
        Ra_d=float('inf')
    else: Ra_d=(beam_dis)*(1+(math.pi*((beam_waist)**2)/((wavelength *(beam_dis))))**2)
    k=2*math.pi/wavelength
    s_overlap=(2/((Wa_d*cavity_waist)*(1/(Wa_d)**4 + 1/(cavity_waist)**4 + 2/((Wa_d)**2*(cavity_waist)**2) + k**2/(4*(Ra_d**2)))**.5))**4  #raise to fourth since we must account for y dimension mismatch too and want intensity not E-field#
    return s_overlap

def frequency_overlap(laser_FWHM=None, cavity_FWHM=None, finesse=None, FSR=1.5*10**9, Rin=None, Rout=None):
    if laser_FWHM==None:
        return 1
    if finesse==None:
        finesse=(math.pi*((Rin*Rout)**.25)/(1-(Rin*Rout)**.5))
    if cavity_FWHM==None:
        cavity_FWHM=FSR/finesse
    f_overlap=cavity_FWHM/laser_FWHM
    if f_overlap >1:
        f_overlap=1
    return f_overlap

def impedance_matching(Rin, Rout):   #slightly less than 1 answers are not rounding errors--answer is the reflected dip as a proportion of the non-resonance signal, which will converge to 1 for Rin=Rout--> inf
    R=Rin*Rout
    max_reflected_intensity=(Rin+(R/Rin)+2*(R**.5))/(1+R+2*(R**.5))
    min_reflected_intensity=(Rin+(R/Rin)-2*(R**.5))/(1+R-2*(R**.5))
    med_reflected_intensity=(Rin+R/Rin)/(1+R)
    peak_to_peak=max_reflected_intensity-min_reflected_intensity
    return peak_to_peak


def cavity_coupling(beam_waist, cavity_waist, beam_dis, Rin, Rout, laser_FWHM, wavelength=1064, FSR=1.5*10**9, cavity_FWHM=None, finesse=None):
    cavity_coupling=spacial_overlap(beam_waist, cavity_waist, beam_dis, wavelength)*(frequency_overlap(laser_FWHM, cavity_FWHM, finesse, FSR, Rin, Rout))
    print('Spatial Coupling = ' + str(round(100*spacial_overlap(beam_waist, cavity_waist, beam_dis, wavelength),3)) + '%')  
    print('Linewidth Coupling = ' + str(round(100*frequency_overlap(laser_FWHM, cavity_FWHM, finesse, FSR),3)) + '%')  
    print('Cavity Coupling = ' + str(round(100*cavity_coupling,3)) + '%')
    

def reflected_dip(beam_waist, cavity_waist, beam_dis, Rin, Rout, laser_FWHM=None, cavity_FWHM=None, finesse=None, FSR=1.5*10**9, wavelength=1064):
    #this output is the percent of the dip in the reflected signal when it is on resonance compared to its maximum (off resonance) value, accounting for spacial coupling, frequency overlap, and impedance matching
    #So, a reflected dip of 70% indicates that when on resonance the reflected signal should drop to 30% of its normal value) 
    reflected_dip=spacial_overlap(beam_waist, cavity_waist, beam_dis, wavelength)*(frequency_overlap(laser_FWHM, cavity_FWHM, finesse, FSR, Rin, Rout))*(impedance_matching(Rin,Rout))
    print('Spatial Coupling = ' + str(round(100*spacial_overlap(beam_waist, cavity_waist, beam_dis, wavelength),3)) + '%')
    print('Linewidth Coupling = ' + str(round(100*frequency_overlap(laser_FWHM, cavity_FWHM, finesse, FSR, Rin, Rout),3)) + '%')
    print('Impedance Matching = ' + str(round(100*impedance_matching(Rin, Rout),1))+'%')
    print('Reflected Dip = ' + str(round(100*reflected_dip,1))+ '% of reflected beam')
